/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"


//
// Constants
//
#define HEIGHTMAP_WIDTH			64
#define HEIGHTMAP_HEIGHT		64

#define TERRAIN_NUMROWSCOLS		64

#define PAINTAREA_WIDTH			150
#define PAINTAREA_HEIGHT		150


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	// initialize extension manager
	if (!GLeeInit())
		return false;

	// initialize time counter
	m_dElapTime = 0;

	// initialize our displacement map
	if( !m_HeightMap.Initialize( HEIGHTMAP_WIDTH, HEIGHTMAP_HEIGHT ) )
		return false;
	m_HeightMap.Refresh();

	// create a patch to represent the terrain
	m_pTerrain = CPrimitiveFactory::MakePatch( 10.f, 10.f, TERRAIN_NUMROWSCOLS, 
		TERRAIN_NUMROWSCOLS );

	if( !m_pTerrain )
		return false;

	// create a quad for the interactive paint surface
	m_pPaintSurface = CPrimitiveFactory::MakeQuad( (float)PAINTAREA_WIDTH, 
		(float)PAINTAREA_HEIGHT );

	if( !m_pPaintSurface )
		return false;

	// initialize the patch
	UpdatePatch();

	return true;
}

bool MGLApp::Shutdown()
{
	CPrimitiveFactory::DeletePrimitive( &m_pTerrain );
	CPrimitiveFactory::DeletePrimitive( &m_pPaintSurface );

	m_HeightMap.Destroy();

	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid divide by zero
	if( height == 0 )
		height = 1;					

	m_nWindowWidth = width;
	m_nWindowHeight = height;

	// set up GL viewport
	glViewport( 0, 0, width, height );
}

void MGLApp::Update( double dt )
{
	// accumulate our time counter
	m_dElapTime = m_dElapTime + dt;
}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	//
	// Draw the terrain patch
	//

	// normal is positive Y
	glNormal3f( 0.f, 1.f, 0.f );

	// set up a rotating perspective view
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)m_nWindowWidth/(GLfloat)m_nWindowHeight, 1.0f, 1000.0f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0.0f, 12.0f, 14.0f, -1.5f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
	glRotatef( 25.f * (float)m_dElapTime, 0.f, 1.f, 0.f );

	// draw the patch
	glDisable( GL_TEXTURE_2D );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_POLYGON_OFFSET_FILL );
	glPolygonOffset( 1.f, 1.f );
	glColor4f( 1.f, 1.f, 1.f, 1.f );
	m_pTerrain->Draw( false, false, true, false );

	// redraw the patch as a black wireframe
	glColor4f( 0.f, 0.f, 0.f, 1.f );
	m_pTerrain->Draw( true, false, false, false );

	//
	// Draw the billboarded height map
	//

	// set up an ortho projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0, (double)m_nWindowWidth, 0, (double)m_nWindowHeight, -1.f, 1.f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	// bind regular (non ATI-float) displacement texture
	m_HeightMap.Bind( FALSE );

	// set up texture state
	glEnable( GL_TEXTURE_2D );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	// draw the quad
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_POLYGON_OFFSET_FILL );
	m_pPaintSurface->Draw( false, false, true, true );

	// overdraw a wireframe outline
	glDisable( GL_TEXTURE_2D );
	glColor4f( 0.5f, 0.5f, 0.5f, 0.5f );
	m_pPaintSurface->Draw( true, false, false, false );
}

void MGLApp::MouseDownPosition( int x, int y )
{
	// convert the y coord from window space to GL screen space
	y = m_nWindowHeight - y;

	if( x < PAINTAREA_WIDTH && y < PAINTAREA_HEIGHT )
	{
		// update the height map
		m_HeightMap.Paint( (float)x/PAINTAREA_WIDTH, (float)y/PAINTAREA_HEIGHT );
		m_HeightMap.Refresh();

		// update the model
		UpdatePatch();
	}
}

void MGLApp::UpdatePatch()
{
	// define some helpers
#define X(p)	(p)[0]
#define Y(p)	(p)[1]
#define Z(p)	(p)[2]
#define R(p)	(p)[0]
#define G(p)	(p)[1]
#define B(p)	(p)[2]
#define A(p)	(p)[3]

	// update the terrain from the height map
	GLfloat *pV = m_pTerrain->pVert;
	GLubyte *pC = m_pTerrain->pCol;

	for( int x=0 ; x<=TERRAIN_NUMROWSCOLS ; x++ )
	{
		for( int y=0 ; y<=TERRAIN_NUMROWSCOLS ; y++ )
		{
			// resolve texture coordinate
			float u = 0.5f+X(pV)/10.f, v = 0.5f+Z(pV)/10.f;

			// displace along Y
			Y(pV) = 5.f * m_HeightMap.GetHeightAt( u, v );

			// set color
			R(pC) = (GLubyte)( pV[1] / 6.f * (float)0xFF );
			B(pC) = 0xFF - R(pC);
			G(pC) = 0;

			// increment pointers
			pV += 3;
			pC += 4;
		}
	}
}
