/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"
#include <stdio.h>
#include <math.h>


//
// Constants
//
#define FILE_VERT_SHADER		"parallax_vert_shad.txt"
#define FILE_FRAG_SHADER		"parallax_frag_shad.txt"

#define FILE_COLOR_TEX			"parallax_color_texture.tga"
#define FILE_NORMBUMP_TEX		"parallax_normbump_texture.tga"

#define TEXTURE_SIZE			512		// square texture

#define EYE_RADIUS				1.8f	// distance from eye to origin


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	// initialize extension manager
	if (!GLeeInit())
		return false;

	// initialize time and frame counters
	m_dElapTime = 0.0;

	// initialize model rotation
	m_fRotTheta = DEG2RAD(-90.f); // rotate the camera a little to begin with
	m_fRotPhi = DEG2RAD(90.f);
	m_fLightX = 50.f; // move scene light a little to begin with
	m_fLightZ = 0.f;

	// initialize a quad primitive
	m_pQuad = CPrimitiveFactory::MakePatch( 1.f, 1.f, 1, 1 );

	// compute tangent data for the quad
	m_pTangentData = new Vector3[ 4 ];
	m_pTangentData[0].Set( 1.f, 0.f, 0.f );
	m_pTangentData[1].Set( 1.f, 0.f, 0.f );
	m_pTangentData[2].Set( 1.f, 0.f, 0.f );
	m_pTangentData[3].Set( 1.f, 0.f, 0.f );
	
	// load color and normal/bump textures
	TargaImage img;

	if( !img.Load( FILE_COLOR_TEX ) )
		return false;

	glGenTextures( 1, &m_hColorTex );
	glBindTexture( GL_TEXTURE_2D, m_hColorTex );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_SIZE, TEXTURE_SIZE, 0, GL_RGBA, GL_UNSIGNED_BYTE, img.GetImage() );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	img.Release();

	if( !img.Load( FILE_NORMBUMP_TEX ) )
		return false;

	glGenTextures( 1, &m_hNormBumpTex );
	glBindTexture( GL_TEXTURE_2D, m_hNormBumpTex );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_SIZE, TEXTURE_SIZE, 0, GL_RGBA, GL_UNSIGNED_BYTE, img.GetImage() );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	img.Release();

	// load shaders
	int iErrTextLen;
	char szErrText[ 128 ];
	
	iErrTextLen = 128;
	m_hVertShad = LoadGLSLShader( GL_VERTEX_SHADER_ARB, FILE_VERT_SHADER, szErrText, &iErrTextLen );
	if( !m_hVertShad || iErrTextLen > 0 )
		return false;

	iErrTextLen = 128;
	m_hFragShad = LoadGLSLShader( GL_FRAGMENT_SHADER_ARB, FILE_FRAG_SHADER, szErrText, &iErrTextLen );
	if( !m_hFragShad || iErrTextLen > 0 )
		return false;

	m_hShadProg = glCreateProgramObjectARB();
	glAttachObjectARB( m_hShadProg, m_hVertShad );
	glAttachObjectARB( m_hShadProg, m_hFragShad );
	glLinkProgramARB( m_hShadProg );

	// enable the shader
	glUseProgramObjectARB( m_hShadProg );

	// attach tangent vector data
	GLint iTangentAttrib = glGetAttribLocationARB( m_hShadProg, "rm_Tangent" );
	glEnableVertexAttribArrayARB( iTangentAttrib );
	glVertexAttribPointerARB( iTangentAttrib, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)m_pTangentData );

	// attach texture samplers
	GLint iColorSampler = glGetUniformLocationARB( m_hShadProg, "TextureColor" );
	glActiveTexture( GL_TEXTURE0 );
	glBindTexture( GL_TEXTURE_2D, m_hColorTex );
	glUniform1iARB( iColorSampler, 0 );

	GLint iNormBumpSampler = glGetUniformLocationARB( m_hShadProg, "TextureNormBump" );
	glActiveTexture( GL_TEXTURE1 );
	glBindTexture( GL_TEXTURE_2D, m_hNormBumpTex );
	glUniform1iARB( iNormBumpSampler, 1 );

	return (glGetError() == GL_NO_ERROR);
}

bool MGLApp::Shutdown()
{
	CPrimitiveFactory::DeletePrimitive( &m_pQuad );
	delete[] m_pTangentData;

	glDeleteObjectARB( m_hVertShad );
	glDeleteObjectARB( m_hFragShad );
	glDeleteObjectARB( m_hShadProg );
	glDeleteTextures( 1, &m_hColorTex );
	glDeleteTextures( 1, &m_hNormBumpTex );

	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid divide by zero
	if( height == 0 )
		height = 1;

	m_nWindowWidth = width;
	m_nWindowHeight = height;

	glViewport( 0, 0, m_nWindowWidth, m_nWindowHeight );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (float)m_nWindowWidth/m_nWindowHeight, 1.0f, 1000.0f );
}

void MGLApp::Update( double dt )
{
	// accumulate our time counter
	m_dElapTime = m_dElapTime + dt;
}

void MGLApp::Render()
{
	// clear the screen
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// use spherical coord conversions to calculate view position
	float fEyeX = EYE_RADIUS*cos(m_fRotTheta)*sin(m_fRotPhi),
		fEyeY = EYE_RADIUS*sin(m_fRotTheta)*sin(m_fRotPhi),
		fEyeZ = EYE_RADIUS*cos(m_fRotPhi);

	// set up interactive view
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( fEyeX, fEyeY, fEyeZ, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f);

	// set shader light position
	GLint iLightPosUniform = glGetUniformLocationARB( m_hShadProg, "v3LightPos" );
	glUniform3fARB( iLightPosUniform, m_fLightX, 7.f, m_fLightZ );

	// set shader view position
	GLint iViewPosUniform = glGetUniformLocationARB( m_hShadProg, "v3ViewPosition" );
	glUniform3fARB( iViewPosUniform, fEyeX, fEyeY, -fEyeZ );

	// cull backface
	glEnable( GL_CULL_FACE );

	// draw the quad
	m_pQuad->Draw( false, true, false, true );
}

void MGLApp::LeftMouseMove( int dx, int dy )
{
	m_fRotTheta += DEG2RAD( (float)dx / 3.f );
	m_fRotPhi += DEG2RAD( (float)dy / 3.f );
}

void MGLApp::RightMouseMove( int dx, int dy )
{
	m_fLightX += (float)dx / 10.f;
	m_fLightZ += (float)dy / 10.f;
}





