/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"
#include <gl/glut.h>


#define DYN_TEX_WIDTH		256
#define DYN_TEX_HEIGHT		256


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::InitPBuffer()
{
	HDC hPrimaryDevCtx = wglGetCurrentDC();

	// Define the minimum requirements we have for the P-Buffer, 
	// and ask WGL to choose a pixel format based on those 
	// requirements.
	int pPixFmtRequirements[] = {
		WGL_DRAW_TO_PBUFFER_ARB, TRUE,      // Enable rendering to buffer
		WGL_BIND_TO_TEXTURE_RGBA_ARB, TRUE, // Usable as a texture
		WGL_SUPPORT_OPENGL_ARB, TRUE,       // Associate with OpenGL
		WGL_DOUBLE_BUFFER_ARB, FALSE,       // Single-buffered
		WGL_RED_BITS_ARB, 8,                // 8 bits of red minimum
		WGL_GREEN_BITS_ARB, 8,              // 8 bits of green minimum 
		WGL_BLUE_BITS_ARB, 8,               // 8 bits of blue minimum 
		WGL_DEPTH_BITS_ARB, 16,             // 16 bits for depth minimum 
		0                                   // End of list
	};

	int iPixelFormat;
	unsigned int iFormatCount;

	wglChoosePixelFormatARB( 
		hPrimaryDevCtx,                     // Device context to query
		(const int*) pPixFmtRequirements,   // List of integer parameters
		NULL,                               // List of float parameters
		1,                                  // Number of formats to return
		&iPixelFormat,                      // The outputted pixel format
		&iFormatCount );                    // Number of outputted formats

	if( iFormatCount == 0 )
		return false;

	// Now that we have a device context and a pixel format, we can 
	// create the P-Buffer.
	int pBufRequirements [] =
	{
		// P-Buffer will bind to a 2D texture
		WGL_TEXTURE_TARGET_ARB, WGL_TEXTURE_2D_ARB,
		// P-Buffer will bind to a texture with format RGBA
		WGL_TEXTURE_FORMAT_ARB, WGL_TEXTURE_RGBA_ARB,
		0
	};

	m_PBuffer.hPBuffer = wglCreatePbufferARB( 
		hPrimaryDevCtx,                     // Window device context
		iPixelFormat,                       // Chosen pixel format
		DYN_TEX_WIDTH,                      // P-Buffer width
		DYN_TEX_HEIGHT,                     // P-Buffer height
		pBufRequirements );                 // No extra attributes

	if( m_PBuffer.hPBuffer == NULL )
		return false;

	m_PBuffer.hDC = wglGetPbufferDCARB( m_PBuffer.hPBuffer );
	m_PBuffer.hGLRC = wglCreateContext( m_PBuffer.hDC );

	return true;
}

bool MGLApp::Init()
{	
	// initialize extension manager
	if (!GLeeInit())
		return false;

	// check for required extensions for hardware support
	if( !WGL_ARB_pbuffer || !WGL_ARB_pixel_format || !WGL_ARB_render_texture )
		return false;

	// initialize time and frame counters
	m_dElapTime = 0.0;
	m_dFPSTime = 0.0;
	m_nFrames = 0;
	m_dFPS = 0.0;

	// init model rotation
	m_fLeftX = 0.f;
	m_fLeftY = 0.f;
	m_fRightX = 0.f;
	m_fRightY = 0.f;

	// save off our window's device and rendering contexts
	m_hDCWnd = wglGetCurrentDC();
	m_hGLRCWnd = wglGetCurrentContext();

	// initialize the p-buffer
	if( InitPBuffer() == false )
		return false;

	return true;
}

bool MGLApp::Shutdown()
{
	// destroy the p-buffer
	if( wglGetCurrentContext() == m_PBuffer.hGLRC )
		wglMakeCurrent( 0, 0 );

	wglDeleteContext( m_PBuffer.hGLRC );
	wglReleasePbufferDCARB( m_PBuffer.hPBuffer, m_PBuffer.hDC );
	wglDestroyPbufferARB( m_PBuffer.hPBuffer );

	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid divide by zero
	if( height == 0 )
		height = 1;					

	m_nWindowWidth = width;
	m_nWindowHeight = height;
}

void MGLApp::Update( double dt )
{
	// accumulate our time counter
	m_dElapTime = m_dElapTime + dt;
}

void MGLApp::DrawDodecahedron()
{
	// this is the object to be rendered to a dynamic texture, so only 
	// render to a small portion of the screen
	glViewport( 0, 0, DYN_TEX_WIDTH, DYN_TEX_HEIGHT );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, 1.0f, 1.0f, 1000.0f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0.0f, 0.0f, 5.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);

	float fHeight = (float)sin(5.0*m_dElapTime)*0.2f;
	glTranslatef( 0.f, fHeight, 0.f );

	glRotatef( m_fRightX/3.f, 0.f, 1.f, 0.f );
	glRotatef( m_fRightY/3.f, 1.f, 0.f, 0.f );

	glutSolidDodecahedron();
}

void MGLApp::DrawTeapot()
{
	// render to the whole screen
	glViewport( 0, 0, m_nWindowWidth, m_nWindowHeight );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)m_nWindowWidth/(GLfloat)m_nWindowHeight, 1.0f, 1000.0f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0.0f, 0.0f, 4.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
	glRotatef( m_fLeftX/3.f, 0.f, 1.f, 0.f );
	glRotatef( m_fLeftY/3.f, 1.f, 0.f, 0.f );

	glutSolidTeapot( 1.f );
}

void MGLApp::Render()
{
	float pfWhite[4] = { 1.f, 1.f, 1.f, 1.f };
	float pfGreen[4] = { 0.5f, 1.f, 0.5f, 1.f };


	//
	// Render a mesh to a texture
	//

	// set the p-buffer as our active render target
	wglMakeCurrent( m_PBuffer.hDC, m_PBuffer.hGLRC );

	// enable lighting in the p-buffer's context to make things 
	// look a little less boring
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, pfWhite );
	glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 50.f );
	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );

	// clear the screen
	glClearColor( 1.f, 0.5f, 0.5f, 1.f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// draw the first object to the window
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, pfGreen );
	glDisable( GL_TEXTURE_2D );
	DrawDodecahedron();


	//
	// Apply the dynamic texture to another object
	//

	// set the window as our active render target
	wglMakeCurrent( m_hDCWnd, m_hGLRCWnd );

	// also enable lighting in the main window's context to 
	// make things look a little less boring
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, pfWhite );
	glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 50.f );
	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );

	// bind the p-buffer to the default texture
	wglBindTexImageARB( m_PBuffer.hPBuffer, WGL_FRONT_LEFT_ARB );

	// clear the screen
	glClearColor( 0.7f, 0.3f, 0.3f, 1.f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// set up texture state for the second pass
	glEnable( GL_TEXTURE_2D );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );

	// draw the second object with the first object textured onto it
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, pfWhite );
	DrawTeapot();

	// release the p-buffer from the texture
	wglReleaseTexImageARB( m_PBuffer.hPBuffer, WGL_FRONT_LEFT_ARB );


	//
	// Draw the framerate
	//

	DrawFPS();
}

void MGLApp::DrawFPS()
{
	// update FPS value
	if( ++m_nFrames == 100 )
	{
		m_dFPS = m_nFrames / (m_dElapTime - m_dFPSTime);
		m_nFrames = 0;
		m_dFPSTime = m_dElapTime;
	}

	// draw the FPS to the screen
	glPushMatrix();
	{
		static char szText[32];

		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glOrtho( 0.f, 100.f, 0.f, 100.f, -100.f, 100.f );

		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();

		glDisable( GL_LIGHTING );
		glDisable( GL_TEXTURE_2D );
		glColor3f( 1.0f, 1.0f, 1.0f );
		glRasterPos2f(0, 0);

		sprintf( szText, "FPS=%.2f", m_dFPS );
		for( char* p = szText ; *p ; p++ )
			glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, *p );
	}
	glPopMatrix();
}

void MGLApp::LeftMouseMove( int dx, int dy )
{
	m_fLeftX += (float)dx;
	m_fLeftY += (float)dy;
}

void MGLApp::RightMouseMove( int dx, int dy )
{
	m_fRightX += (float)dx;
	m_fRightY += (float)dy;
}

