#include "displacementmap.h"


// helpers
#define CLAMPED(x)			((x)<0.f ? 0.f : (x)>1.f ? 1.f : (x))
#define IS_WITHIN(x,l,r)	(((x) >= l) && ((x) <= r))


// brush used to modify the height map
static const float	BRUSH_LUMINANCE = 0.8f;		// range from 0.f to 25.f
static const int	BRUSH_WIDTH = 9;
static const int	BRUSH_HEIGHT = 9;
static const char	BRUSH[ BRUSH_WIDTH*BRUSH_HEIGHT+1 ] = 
"\
003333300\
033555330\
335575533\
355797553\
357999753\
355797553\
335575533\
033555330\
003333300\
";


//////////////////////////////////////////////////////////////////////////
// CDisplacementMap class implementation
//////////////////////////////////////////////////////////////////////////
CDisplacementMap::CDisplacementMap()
{
	m_pPixels = NULL;
	m_nStdTexID = 0;
	m_nATITexID = 0;
}

CDisplacementMap::~CDisplacementMap()
{
	Destroy();
}

bool CDisplacementMap::Initialize( int width, int height )
{
	// initialize the texture buffer
	m_nWidth = width;
	m_nHeight = height;
	m_nImgSize = width * height * sizeof(GLfloat);

	m_pPixels = new GLfloat[ width * height ];
	memset( m_pPixels, 0, m_nImgSize );

	// generate a GL name for this texture
	glGenTextures( 1, &m_nStdTexID );

	if( GLEE_ATI_texture_float )
		glGenTextures( 1, &m_nATITexID );

	return true;
}

void CDisplacementMap::Refresh()
{
	// update GL's internal copy of the texture

	// standard float format
	Bind( false );
	glTexImage2D( GL_TEXTURE_2D, 0, 1, m_nWidth, m_nHeight, 
		0, GL_LUMINANCE, GL_FLOAT, m_pPixels );

	// ATI float format
	// Note: renders prohibitively slow when not used as vertex texture
	if( GLEE_ATI_texture_float )
	{
		Bind( true );
		glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE_FLOAT32_ATI, 
			m_nWidth, m_nHeight, 0, GL_LUMINANCE, GL_FLOAT, m_pPixels );
	}
}

void CDisplacementMap::Destroy()
{
	// free GL's copies of the texture
	if( m_nStdTexID )
	{
		glDeleteTextures( 1, &m_nStdTexID );
		m_nStdTexID = 0;
	}

	if( GLEE_ATI_texture_float && m_nATITexID )
	{
		glDeleteTextures( 1, &m_nATITexID );
		m_nATITexID = 0;
	}

	// delete the local copy
	if( m_pPixels )
	{
		delete[] m_pPixels;
		m_pPixels = NULL;
	}
}

void CDisplacementMap::Bind( bool bUseATIFloat )
{
	// check parameter validity
	if( !GLEE_ATI_texture_float && bUseATIFloat )
		return;

	// set our texture as active
	glBindTexture( GL_TEXTURE_2D, bUseATIFloat ? m_nATITexID : m_nStdTexID );
}

void CDisplacementMap::Paint( float u, float v )
{
	// find the paint location on the destination texture
	int startX = (int)(CLAMPED(u) * (m_nWidth-1)) - BRUSH_WIDTH/2, 
		startY = (int)(CLAMPED(v) * (m_nHeight-1)) - BRUSH_HEIGHT/2;

	// additively blend the brush bitmap with the texture
	for( int y=0 ; y<BRUSH_HEIGHT ; y++ )
	{
		for( int x=0 ; x<BRUSH_WIDTH ; x++ )
		{
			// calculate position on texture
			int texX = startX + x, texY = startY + y;

			// guard against scanline overrun/underrun
			if( IS_WITHIN( texX, 0, m_nWidth-1 ) && 
				IS_WITHIN( texY, 0, m_nHeight-1 ) )
			{
				int iBrushVal = (int)(BRUSH_LUMINANCE * (int)(BRUSH[x+y*BRUSH_WIDTH] - '0'));
				GLfloat *pTexel = &m_pPixels[texX+texY*m_nWidth];
				GLfloat fNewLuminance = (*pTexel) + (GLfloat)iBrushVal / 256.f;

				// clamp updated luminance
				fNewLuminance = CLAMPED( fNewLuminance );

				(*pTexel) = fNewLuminance;
			}
		}
	}
}

GLfloat CDisplacementMap::GetHeightAt( float u, float v )
{
	int texX = (int)(CLAMPED(u) * (m_nWidth-1)), 
		texY = (int)(CLAMPED(v) * (m_nHeight-1));

	return (float)m_pPixels[texX+texY*m_nWidth];
}
