#include "primitivefactory.h"


// helpers
#define SETVERT(p,x,y,z) \
	{ (p)[0]=(x); (p)[1]=(y); (p)[2]=(z); (p)+=3;				}

#define SETCOL(p,r,g,b,a) \
	{ (p)[0]=(r); (p)[1]=(g); (p)[2]=(b); (p)[3]=(a); (p)+=4;	}

#define SETNORMAL(p,x,y,z) \
	{ (p)[0]=(x); (p)[1]=(y); (p)[2]=(z); (p)+=3;				}

#define SETTEXCOORD(p,u,v) \
	{ (p)[0]=(u); (p)[1]=(v); (p)+=2;							}

#define SETTRI(p,i1,i2,i3) \
	{ (p)[0]=(i1); (p)[1]=(i2); (p)[2]=(i3); (p) += 3;			}


//////////////////////////////////////////////////////////////////////////
// Primitive struct implementation
//////////////////////////////////////////////////////////////////////////
Primitive::Primitive( GLenum nType, int nVerts, int nIdxs )
	: nElemType( nType )
	, nVertices( nVerts )
	, nIndices( nIdxs )
	, pVert( new GLfloat[3*nVerts] )
	, pCol( new GLubyte[4*nVerts] )
	, pNormal( new GLfloat[3*nVerts] )
	, pTexCoord( new GLfloat[2*nVerts] )
	, pIdx( new GLushort[nIdxs] )
{
	memset( pVert, 0, 3*nVerts*sizeof(GLfloat) );
	memset( pCol, 0, 4*nVerts*sizeof(GLubyte) );
	memset( pNormal, 0, 3*nVerts*sizeof(GLfloat) );
	memset( pTexCoord, 0, 2*nVerts*sizeof(GLfloat) );
	memset( pIdx, 0, nIdxs*sizeof(GLushort) );
}

Primitive::~Primitive()
{
	if( pVert )		delete[] pVert;
	if( pCol )		delete[] pCol;
	if( pNormal )	delete[] pNormal;
	if( pTexCoord )	delete[] pTexCoord;
	if( pIdx )		delete[] pIdx;
}

void Primitive::Draw( bool bWireframe, bool bNormalArray, bool bColorArray, 
                      bool bTexCoordArray )
{
	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_COLOR_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

	if( pVert )
	{
		glEnableClientState( GL_VERTEX_ARRAY );
		glVertexPointer( 3, GL_FLOAT, 0, pVert );
	}

	if( pCol && bColorArray )
	{
		glEnableClientState( GL_COLOR_ARRAY );
		glColorPointer( 4, GL_UNSIGNED_BYTE, 0, pCol );
	}

	if( pNormal && bNormalArray )
	{
		glEnableClientState( GL_NORMAL_ARRAY );
		glNormalPointer( GL_FLOAT, 0, pNormal );
	}

	if( pTexCoord && bTexCoordArray )
	{
		glEnableClientState( GL_TEXTURE_COORD_ARRAY );
		glTexCoordPointer( 2, GL_FLOAT, 0, pTexCoord );
	}

	if( bWireframe )
		glDrawElements( (nElemType==GL_TRIANGLES) ? GL_LINES : GL_LINE_STRIP, 
			nIndices, GL_UNSIGNED_SHORT, pIdx );
	else
		glDrawElements( nElemType, nIndices, GL_UNSIGNED_SHORT, pIdx );
}


//////////////////////////////////////////////////////////////////////////
// CPrimitiveFactory class implementation
//////////////////////////////////////////////////////////////////////////
void CPrimitiveFactory::DeletePrimitive( Primitive **ppPrim )
{
	if( ppPrim && *ppPrim )
	{
		delete (*ppPrim);
		(*ppPrim) = NULL;
	}
}

Primitive* CPrimitiveFactory::MakePatch( float fWidth, float fHeight, 
                                         int nCols, int nRows )
{
	if( fWidth < 0.f || fHeight < 0.f || nCols < 1 || nRows < 1 )
		return NULL;
	
	int idx = 0;

	// calculate number of vertices and indices needed to describe the patch
	int nVerts = (nCols+1)*(nRows+1);
	int nIdxs = nCols*nRows*2*3;

	// initialize the model
	Primitive *pPrim = new Primitive( GL_TRIANGLES, nVerts, nIdxs );

	// make a copy to use as an iterator
	Primitive temp = *pPrim;

	// loop over all patch vertices
	float u = 0.f;
	for( int r=0 ; r<=nRows ; r++, u += 1.f/nRows )
	{
		float v = 0.f;
		for( int c=0 ; c<=nCols ; c++, v += 1.f/nCols )
		{
			// set this vertex
			SETVERT( temp.pVert, (u-0.5f) * fWidth, 0.f, (v-0.5f) * fHeight )
			SETCOL( temp.pCol, 0x00, 0x00, 0x00, 0x00 );
			SETNORMAL( temp.pNormal, 0.f, 1.f, 0.f );
			SETTEXCOORD( temp.pTexCoord, u, v );

			// form the two triangles for the quad at this row and column
			if( c < nCols && r < nRows )
			{
				SETTRI( temp.pIdx, idx, idx+nCols+1, idx+1 );
				SETTRI( temp.pIdx, idx+1, idx+nCols+1, idx+nCols+2 );
			}

			// increment the index counter
			idx++;
		}
	}
	
	// prevent a memory error in temp's destructor
	memset( &temp, 0, sizeof(Primitive) );

	return pPrim;
}

Primitive* CPrimitiveFactory::MakeQuad( float fWidth, float fHeight )
{
	if( fWidth < 0.f || fHeight < 0.f )
		return NULL;

	Primitive *pPrim = new Primitive( GL_TRIANGLE_FAN, 4, 4 );

	pPrim->pVert[3] = pPrim->pVert[6] = fWidth;
	pPrim->pVert[7] = pPrim->pVert[10] = fHeight;

	pPrim->pNormal[1] = pPrim->pNormal[4] = pPrim->pNormal[7] = pPrim->pNormal[10] = 1.f;
	pPrim->pTexCoord[2] = pPrim->pTexCoord[4] = pPrim->pTexCoord[5] = pPrim->pTexCoord[7] = 1.f;

	pPrim->pIdx[0] = 0;
	pPrim->pIdx[1] = 1;
	pPrim->pIdx[2] = 2;
	pPrim->pIdx[3] = 3;

	pPrim->pCol[0] = 0xFF;
	pPrim->pCol[5] = 0xFF;
	pPrim->pCol[10] = 0xFF;
	pPrim->pCol[12] = 0xFF;
	pPrim->pCol[14] = 0xFF;

	return pPrim;
}